/*
 *  Initialization Routines
 *  by Tom Epperly
 *  Created: 3/24/1990
 *  Version: $Revision: 1.26 $
 *  Version control file: $RCSfile: initialize.c,v $
 *  Date last modified: $Date: 97/07/28 20:52:09 $
 *  Last modified by: $Author: ballan $
 *
 *  This file is part of the Ascend Language Interpreter.
 *
 *  Copyright (C) 1990, 1993, 1994 Thomas Guthrie Epperly
 *
 *  The Ascend Language Interpreter is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  The Ascend Language Interpreter is distributed in hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139 USA.  Check the file named
 *  COPYING.
 *
 */

#include "utilities/ascConfig.h"
#include "general/list.h"
#include "general/dstring.h"
#include "compiler/compiler.h"
#include "compiler/symtab.h"
#include "compiler/fractions.h"
#include "compiler/dimen.h"
#include "compiler/functype.h"
#include "compiler/types.h"
#include "compiler/forvars.h"
#include "compiler/name.h"
#include "compiler/find.h"
#include "compiler/vlist.h"
#include "compiler/instance_enum.h"
#include "compiler/stattypes.h"
#include "compiler/statement.h"
#include "compiler/statio.h"
#include "compiler/switch.h"
#include "compiler/module.h"
#include "compiler/evaluate.h"
#include "compiler/value_type.h"
#include "compiler/setinstval.h"
#include "compiler/extfunc.h"
#include "compiler/packages.h"
#include "compiler/instance_io.h"
#include "compiler/nameio.h"
#include "compiler/atomvalue.h"
#include "compiler/instquery.h"
#include "compiler/slist.h"
#include "compiler/proc.h"
#include "compiler/child.h"
#include "compiler/type_desc.h"
#include "compiler/library.h"
#include "compiler/extcall.h"
#include "compiler/initialize.h"
#include "compiler/switch.h"
#include "compiler/exprs.h"
#include "compiler/sets.h"


#ifndef lint
static CONST char InitializeRCSid[]="$Id: initialize.c,v 1.26 97/07/28 20:52:09 ballan Exp $";
#endif /* lint */


/*********************************************************************\
  There is a stack of procedure calls kept for tracing and breaking
  recursion errors.
  INITSTACKLIMIT is the minimum we will allow internally.
\*********************************************************************/

static struct {
  unsigned long limit;
  unsigned long depth;
} g_proc = {INITSTACKLIMIT,0L};

unsigned long GetProcStackLimit(void)
{
  return g_proc.limit;
}

void SetProcStackLimit(unsigned long lim)
{
  if (lim < 3) {
    FPRINTF(ASCERR,
      "SetProcStackLimit called with limit too small (%lu). Ignored.\n",lim);
    return;
  }
  if (g_proc.depth) {
    FPRINTF(ASCERR, "SetProcStackLimit called during evaluation. Ignored.\n");
    return;
  }
  g_proc.limit = lim;
  return;
}

/* The following 2 forward declarations have been moved out of the
 * header, where they had no business being, so we can adequately
 * guard against recursive functions.
 */
static enum Proc_enum ExecuteInitRun(struct Instance *, struct Statement *);
/*********************************************************************\
void ExecuteInitRun(i,stat);
struct Instance *i;
struct InitProcedure *proc;
This will execute a run statement, using the given instance as the
context. stat must be a RUN statement. In the event of error will
print appropriate messages to stderr.
\*********************************************************************/

static enum Proc_enum ExecuteInitProcedure(struct Instance *,
                                           struct InitProcedure *);
/*********************************************************************\
void ExecuteInitProcedure(i,proc)
struct Instance *i;
struct InitProcedure *proc;
This will execute proc on the instance i.
\*********************************************************************/

static enum Proc_enum ExecuteInitStatements(struct Instance *,
                                            struct StatementList *);
static enum Proc_enum ClassAccessRealInitialize(struct Instance *,
                                                struct Name *, struct Name *);
static enum Proc_enum RealInitialize(struct Instance *, struct Name *);

/* just forward declarations cause we need it */

static void InstanceNamePart(struct Name *n, struct Name **copy,
                             CONST char **procname)
{
  register struct Name *ptr,*tmp;
  if (n==NULL){
    *copy = NULL;
    *procname = NULL;
  }
  else if (NextName(n)==NULL){	/* RUN a; a is the procname */
    *copy = NULL;
    if (NameId(n)) *procname = NameIdPtr(n);
    else *procname = NULL;
  }
  else{				/* RUN a.b.c.clear; clear is the procname */
   ptr = *copy = CopyName(n);
   while (NextName(NextName(ptr))!=NULL)
     ptr = NextName(ptr);
   tmp = NextName(ptr);
   LinkNames(ptr,NULL);		/* disconnect last part of name */
   if (NameId(tmp)) *procname = NameIdPtr(tmp);
   else *procname = NULL;
   DestroyName(tmp);
  }
}

static struct InitProcedure *SearchProcList(CONST struct gl_list_t *l,
				     CONST char *name)
{
  register unsigned up,c,low;
  register struct InitProcedure *ptr;
  register int cmp;
  up = gl_length(l);
  low = 1;
  while(low<=up){
    c = (low+up)/2;
    ptr = (struct InitProcedure *)gl_fetch(l,c);
    if ((cmp = strcmp(ProcName(ptr),name))==0) {
      return ptr;
    }
    if (cmp<0) {
      low = c+1;
    } else {
      up = c-1;
    }
  }
  return NULL;
}

struct InitProcedure *FindProcedure(CONST struct Instance *i,
				    CONST char *procname)
{
  struct TypeDescription *desc;
  struct gl_list_t *plist;
  desc = InstanceTypeDesc(i);
  plist = GetInitializationList(desc);
  if (plist){
    return SearchProcList(plist,procname);
  }
  else return NULL;
}


static void WriteInitErr(struct Statement *stat, CONST char *str)
{
  FPRINTF(ASCERR,"METHOD EXECUTION ERROR\n");
  WSEM(ASCERR,stat,str);
}

/*
 * This returns proc_all_ok in all circumstances except stack overflow.
 * If within it any other error occurs, it prints the message and
 * then pretends everything is ok.
 * This behavior should perhaps be better.
 */
static enum Proc_enum ExecuteInitRun(struct Instance *i,
                                     struct Statement *stat)
{
  struct Name *typename;
  enum Proc_enum return_value;

  typename = RunStatAccess(stat);
  if (typename!=NULL) {
    return_value = ClassAccessRealInitialize(i,typename,RunStatName(stat));
  } else {
    return_value = RealInitialize(i,RunStatName(stat));
  }
  if (return_value == Proc_all_ok) return return_value;

  /* an error was encountered */
  WriteInitErr(stat,"RUN statement error");
  switch(return_value) {
  case Proc_bad_name:
    FPRINTF(ASCERR,"Error: bad procedure name in RUN statement\n");
    break;
  case Proc_proc_not_found:
    FPRINTF(ASCERR,"Error: procedure not found in RUN statement\n");
    break;
  case Proc_illegal_name_use:
    FPRINTF(ASCERR,"Error: illegal name use in RUN statement\n");
    break;
  case Proc_type_not_found:
    FPRINTF(ASCERR,"Error: type name not found in RUN statement\n");
    break;
  case Proc_illegal_type_use:
    FPRINTF(ASCERR,"Error: unconformable types in RUN statement\n");
    break;
  case Proc_stack_exceeded_this_frame:
  case Proc_stack_exceeded:
    return return_value;
  default:
    FPRINTF(ASCERR,"Error: type name not found in RUN statement\n");
    break;
  }
  return Proc_all_ok;
}


/*
 * The following functions have been made static as they are very
 * similar to those used in instantiate.c. They really should be
 * rationalized and exported by instantiate.c
 */
#define SELF_NAME "SELF"

static int SpecialSelfName(CONST struct Name *n)
{
  CONST char *id;
  if (n==NULL) {
    return 0;
  }
  id = SimpleNameIdPtr(n);
  if (id==NULL) {
    return 0;
  }
  if (strcmp(id,SELF_NAME)==0) {
    return 1;
  } else {
    return 0;
  }
}

static struct gl_list_t *ProcessArgs(struct Instance *inst,
				     CONST struct VariableList *vl,
				     enum find_errors *ferr)
{
  struct gl_list_t *arglist;
  struct gl_list_t *branch;
  CONST struct Name *n;

  ListMode=1;
  arglist = gl_create(10L);
  while(vl!=NULL){
    n = NamePointer(vl);
    branch = FindInstances(inst,n,ferr);
    if (branch==NULL){
      if (!SpecialSelfName(n)){		/* check for SELF only if it fails */
	DestroySpecialList(arglist);
	ListMode=0;
	return NULL;
      }
      else{
	branch = gl_create(1L);
        /* Self referential instance */
	gl_append_ptr(branch,(VOIDPTR)inst);
      }
    }
    gl_append_ptr(arglist,(VOIDPTR)branch);
    vl = NextVariableNode(vl);
  }
  ListMode=0;
  return arglist;
}

static struct gl_list_t *CheckExtCallArgs(struct Instance *inst,
					  struct Statement *stat,
					  enum find_errors *ferr)
{
  struct VariableList *vl;
  struct gl_list_t *result;

  vl = ExternalStatVlist(stat);
  result = ProcessArgs(inst,vl,ferr);
  if (result==NULL){
    return NULL;
  }
  return result;
}

static
enum Proc_enum ExecuteInitCall(struct Instance *i, struct Statement *stat)
{
  (void)i;     /*  stop gcc whine about unused parameter  */
  (void)stat;  /*  stop gcc whine about unused parameter  */
#if 0
  struct ExternalFunc *efunc;
  CONST char *funcname;
  struct gl_list_t *arglist=NULL;
  enum find_errors ferr;

  funcname = ExternalStatFuncName(stat);
  efunc = LookupExtFunc(funcname);
  if (!efunc) {
    WriteInitErr(stat,"External function has not been loaded");
    return Proc_all_ok;
  }
  eval_func = GetValueFunc(efunc);
  if (!eval_func) {
    WriteInitErr(stat,"Nonexistent Evaluation routines for external function");
    return Proc_all_ok;
  }
  arglist = CheckExtCallArgs(i,stat,&ferr);
  if (ferr != correct_instance) { /* should give more diagnostics here */
    WriteInitErr(stat,"Incorrect arguements to external function.");
/* KHACK: NEED A WRITE FIND ERROR MESSAGE FUNCTION...PUT IN FIND.H
   switch(ferr)
   { case }
*/
    if (arglist) {
      DestroySpecialList(arglist);
    }
    return Proc_all_ok;
  }
  Init_Slv_Interp(&slv_interp);
  nok = (*eval_func)(&slv_interp,i,arglist);
  if (nok) {
    WriteInitErr(stat,"Error in evaluating external function.");
  }
  if (arglist) {
    DestroySpecialList(arglist);
  }
#endif
  return Proc_all_ok;
}

/*
 * This always returns ok. at least as of 5/96
 */
static
enum Proc_enum ExecuteInitExt(struct Instance *i,
                              struct Statement *stat)
{
  struct ExternalFunc *efunc;
  CONST char *funcname;
  struct gl_list_t *arglist=NULL;
  enum find_errors ferr;

  struct Slv_Interp slv_interp;
  int (*eval_func) ();
  int nok;

  funcname = ExternalStatFuncName(stat);
  efunc = LookupExtFunc(funcname);
  if (!efunc) {
    WriteInitErr(stat,"External function has not been loaded");
    return Proc_all_ok;
  }
  eval_func = GetValueFunc(efunc);
  if (!eval_func) {
    WriteInitErr(stat,"Nonexistent Evaluation routines for external function");
    return Proc_all_ok;
  }
  arglist = CheckExtCallArgs(i,stat,&ferr);
  if (ferr != correct_instance) { /* should give more diagnostics here */
    WriteInitErr(stat,"Incorrect arguements to external function.");
/* KHACK: NEED A WRITE FIND ERROR MESSAGE FUNCTION...PUT IN FIND.H
   switch(ferr)
   { case }
*/
    if (arglist) {
      DestroySpecialList(arglist);
    }
    return Proc_all_ok;
  }
  Init_Slv_Interp(&slv_interp);
  nok = (*eval_func)(&slv_interp,i,arglist);
  if (nok) {
    WriteInitErr(stat,"Error in evaluating external function.");
  }
  if (arglist) {
    DestroySpecialList(arglist);
  }
  return Proc_all_ok;
}

/*
 * executes a for loop until some untrapped error occurs.
 * if no untrapped errors occur, returns Proc_all_ok.
 * As of 5/96 the only untrapped errors should be stack overflow.
 */
static enum Proc_enum ExecuteInitFor(struct Instance *i,
                                     struct Statement *stat)
{
  CONST char *name;
  struct Expr *ex;
  struct StatementList *sl;
  unsigned long c,len;
  struct value_t value;
  struct set_t *sptr;
  struct for_var_t *fv;
  enum Proc_enum rval = Proc_all_ok;

  name = ForStatIndex(stat);
  ex = ForStatExpr(stat);
  sl = ForStatStmts(stat);
  if (FindForVar(EvaluationForTable,name)){ /* duplicated for variable */
    WriteInitErr(stat,"FOR construct uses duplicate index variable");
    return rval;  /* header says we're not returning trapped errors */
  }
  assert(!EvaluationContext);
  EvaluationContext = i;
  value = EvaluateExpr(ex,NULL,InstanceEvaluateName);
  EvaluationContext = NULL;
  switch(ValueKind(value)){
  case error_value:
    WriteInitErr(stat,"Error evaluating FOR index values");
    break;
  case set_value:
    sptr = SetValue(value);
    switch(SetKind(sptr)){
    case empty_set: break;
    case integer_set:
      fv = CreateForVar(name);
      SetForVarType(fv,f_integer);
      AddLoopVariable(EvaluationForTable,fv);
      len = Cardinality(sptr);
      switch(ForLoopOrder(stat)){
      case f_random:
        /* fall through, that should never occur due to parser. */
      case f_increasing:
	for(c = 1; c <= len && rval == Proc_all_ok; c++){
	  SetForInteger(fv,FetchIntMember(sptr,c));
	  rval = ExecuteInitStatements(i,sl);
	}
	break;
      case f_decreasing:
	for(c = len; c >= 1 && rval == Proc_all_ok; c--){
	  SetForInteger(fv,FetchIntMember(sptr,c));
	  rval = ExecuteInitStatements(i,sl);
	}
	break;
      }
      RemoveForVariable(EvaluationForTable);
      break;
    case string_set:
      fv = CreateForVar(name);
      SetForVarType(fv,f_symbol);
      AddLoopVariable(EvaluationForTable,fv);
      len = Cardinality(sptr);
      switch(ForLoopOrder(stat)){
      case f_random:
        /* fall through, that should never occur due to parser. */
      case f_increasing:
	for(c = 1; c <= len && rval == Proc_all_ok; c++){
	  SetForSymbol(fv,FetchStrMember(sptr,c));
	  rval = ExecuteInitStatements(i,sl);
	}
	break;
      case f_decreasing:
	for(c = len; c >= 1 && rval == Proc_all_ok; c--){
	  SetForSymbol(fv,FetchStrMember(sptr,c));
	  rval = ExecuteInitStatements(i,sl);
	}
	break;
      }
      RemoveForVariable(EvaluationForTable);
      break;
    }
    break;
  default:
    WriteInitErr(stat,"FOR index expression is not a set expression");
    break;
  }
  DestroyValue(&value);
  return rval;
}

/*
 * returns the result of the statements inside the if given.
 * at present the only non-all_ok return should be stack overflow.
 * 5/96 baa
 */
static enum Proc_enum ExecuteInitIf(struct Instance *i, struct Statement *stat)
{
  struct value_t value;
  enum Proc_enum rval = Proc_all_ok;

  assert(!EvaluationContext);
  EvaluationContext = i;
  value = EvaluateExpr(IfStatExpr(stat),NULL,InstanceEvaluateName);
  EvaluationContext = NULL;
  switch(ValueKind(value)){
  case boolean_value:
    if (BooleanValue(value)) {
      rval = ExecuteInitStatements(i,IfStatThen(stat));
    } else {
      if (IfStatElse(stat)) {
        rval = ExecuteInitStatements(i,IfStatElse(stat));
      }
    }
    break;
  case error_value:
    WriteInitErr(stat,"Error in evaluating IF expression");
    break;
  default:
    WriteInitErr(stat,"IF expression is not a boolean expression");
    break;
  }
  DestroyValue(&value);
  return rval;
}


/*
 * Compare current values of the switching variables with
 * the set of values in a CASE of a SWITCH statement, and try to find
 * is such values are the same. If they are, the function will return 1,
 * else, it will return 0.
 */
static
int AnalyzeSwitchCase(struct Instance *ref, struct VariableList *vlist,
		      struct Set *s)
{
  CONST struct Expr *expr;
  CONST struct Name *name;
  CONST char *value;
  CONST char *symvar;
  CONST struct VariableList *vl;
  CONST struct Set *values;
  int val;
  int valvar;
  struct gl_list_t *instances;
  struct Instance *inst;
  enum find_errors err;
  symchar *str;
  struct for_var_t *fvp;

  assert(s!= NULL);
  assert(vlist != NULL);
  values = s;
  vl = vlist;

  while (vl!=NULL) {
    name = NamePointer(vl);
    expr = GetSingleExpr(values);
    instances = FindInstances(ref,name,&err);
    if (instances == NULL){
      FPRINTF(stderr,
          "No instance found for a variable name in a SWITCH statement  \n");
      return 0;
    }
    else {
      if (gl_length(instances)==1) {
        inst = (struct Instance *)gl_fetch(instances,1);
        gl_destroy(instances);
        switch(ExprType(expr)) {
          case e_boolean:
	    val =  ExprBValue(expr);
	    if (val == 2) { /* ANY */
	      break;
	    }
	    valvar = GetBooleanAtomValue(inst);
	    if (val != valvar) return 0;
	    break;
          case e_int:
	    val =  ExprIValue(expr);
	    valvar = GetIntegerAtomValue(inst);
	    if (val != valvar) return 0;
	    break;
          case e_symbol:
	    symvar = ExprSymValue(expr);
	    value = GetSymbolAtomValue(inst);
	    if (strcmp(symvar,value)) {
              return 0;
            }
	    break;
          case e_var:
            if (EvaluationForTable&&
               (str = SimpleNameIdPtr(ExprName(expr)))&&
               (fvp=FindForVar(EvaluationForTable,str))) {
              switch (GetForKind(fvp)) {
                case f_integer:
	          val = GetForInteger(fvp);
	          valvar = GetIntegerAtomValue(inst);
	          if (val != valvar) return 0;
	          break;
                case f_symbol:
	          symvar = GetForSymbol(fvp);
	          value = GetSymbolAtomValue(inst);
	          if (strcmp(symvar,value)) {
                    return 0;
                  }
	          break;
	        default:
	          FPRINTF(stderr,
                        "Wrong index expression in AnalyzeSwitchCase \n");
                  return 0;
	      }
	    } else {
	      FPRINTF(stderr,"Wrong index expression in AnalyzeSwitchCase \n");
              return 0;
	    }
            break;
          default:
	    FPRINTF(stderr,"Wrong value expression in AnalyzeSwitchCase \n");
            return 0;
        }
      } else {
        gl_destroy(instances);
	FPRINTF(stderr,
            "Duplicate instances for a variable name in a SWITCH statement\n");
        return 0;
      }
    }
    vl = NextVariableNode(vl);
    values = NextSet(values);
  }
  return 1;
}


/* This function will determine which case of a SWITCH statement
 * applies for the current values of the switching variables.
 * this function  will call for the execution of the cases which
 * match. It handles OTHERWISE properly (case when set == NULL).
 */

static
enum Proc_enum ExecuteInitSwitch(struct Instance *i,
		                 struct Statement *stat)
{
  struct VariableList *vlist;
  struct SwitchList *sw;
  struct Set *set;
  struct StatementList *sl;
  int case_match;
  enum Proc_enum rval = Proc_all_ok;

  vlist = SwitchStatVL(stat);
  sw = SwitchStatCases(stat);
  case_match =0;

  while (sw!=NULL){
    set = SwitchSetList(sw);
    sl = SwitchStatementList(sw);
    if (set != NULL) {
      if (AnalyzeSwitchCase(i,vlist,set)) {
        rval = ExecuteInitStatements(i,sl);
        case_match++;
        if (rval != Proc_all_ok) {
          WriteInitErr(stat,"Error in execution of SWITCH statement\n");
          return rval;
	}
      }
    }
    else {
      if (!case_match) {
        rval = ExecuteInitStatements(i,sl);
        case_match = 1;
        if (rval != Proc_all_ok) {
          WriteInitErr(stat,"Error in execution of SWITCH statement\n");
          return rval;
	}
      }
    }
    sw = NextSwitchCase(sw);
  }
  if (case_match == 0) {
    FPRINTF(ASCERR,"No case matched in SWITCH statement\n");
  }
  return rval;
}


static void AssignInitValue(struct Instance *i, struct value_t v,
		     struct Statement *stat)
{
  CONST dim_type *dim;
  switch(InstanceKind(i)) {
  case MODEL_INST:
  case ARRAY_INT_INST:
  case ARRAY_ENUM_INST:
  case REL_INST:
  case DUMMY_INST:
    WriteInitErr(stat,"Assignment to a non-atomic instance");
    break;
  case INTEGER_INST:
  case INTEGER_ATOM_INST:
    SetIntegerAtomValue(i,IntegerValue(v),0);
    break;
  case SET_INST:
  case SET_ATOM_INST:
  case REAL_CONSTANT_INST:
  case BOOLEAN_CONSTANT_INST:
  case INTEGER_CONSTANT_INST:
  case SYMBOL_CONSTANT_INST:
    WriteInitErr(stat, "Assignment to a set or constant instance");
    break;
  case REAL_INST:
  case REAL_ATOM_INST:
    switch(ValueKind(v)){
    case real_value:
      dim = CheckDimensionsMatch(RealValueDimensions(v),RealAtomDims(i));
      if (dim==NULL){
	WriteInitErr(stat,"Dimensionally inconsistent assignment");
      } else {
	if (dim!=RealAtomDims(i)) SetRealAtomDims(i,dim);
	SetRealAtomValue(i,RealValue(v),0);
      }
      break;
    case integer_value:
      dim = CheckDimensionsMatch(Dimensionless(),RealAtomDims(i));
      if (dim==NULL){
	WriteInitErr(stat,"Dimensionally inconsistent assignment");
      }
      else{
	if (dim!=RealAtomDims(i)) SetRealAtomDims(i,dim);
	SetRealAtomValue(i,(double)IntegerValue(v),0);
      }
      break;
    default:
      WriteInitErr(stat,
		   "Right hand side of assignment is not a real expression"
		   );
      break;
    }
    break;
  case BOOLEAN_INST:
  case BOOLEAN_ATOM_INST:
    if (ValueKind(v)!=boolean_value){
      WriteInitErr(stat,"Right hand side of assignment is not a boolean");
    } else {
      SetBooleanAtomValue(i,BooleanValue(v),0);
    }
    break;
  case SYMBOL_INST:
  case SYMBOL_ATOM_INST:
    if (ValueKind(v)!=symbol_value){
      WriteInitErr(stat,"Right hand side of assignment is not a symbol");
    } else {
      SetSymbolAtomValue(i,SymbolValue(v));
    }
    break;
  default:
    WriteInitErr(stat,"Assignment to bogus instance type.");
    break;
  }
}

/* this function always returns ok. 5/96 */
static enum Proc_enum ExecuteInitAsgn(struct Instance *i,
                                      struct Statement *stat)
{
  struct gl_list_t *instances;
  struct Instance *inst;
  unsigned c,len;
  struct value_t value;
  enum find_errors err;

  instances = FindInstances(i,DefaultStatVar(stat),&err);
  if (instances != NULL){
    assert(!EvaluationContext);
    EvaluationContext = i;
    value = EvaluateExpr(DefaultStatRHS(stat),NULL,InstanceEvaluateName);
    EvaluationContext = NULL;
    if (ValueKind(value)==error_value) {
      WriteInitErr(stat,"Error evaluating assignment right hand side");
    } else {
      len = gl_length(instances);
      for(c=1;c<=len;c++){
	inst = (struct Instance *)gl_fetch(instances,c);
	AssignInitValue(inst,value,stat);
      }
    }
    DestroyValue(&value);
    gl_destroy(instances); /* oops, left this out */
  }
  else{				/* error finding left hand side */
    WriteInitErr(stat,"Error finding left hand side instance");
  }
  return Proc_all_ok;
}

/* will return all_ok unless stack overflows */
static enum Proc_enum ExecuteInitStatement(struct Instance *i,
                                           struct Statement *stat)
{
  switch(StatementType(stat)){
  case FOR:
    return ExecuteInitFor(i,stat);
  case ASGN:
    return ExecuteInitAsgn(i,stat);
  case RUN:
    return ExecuteInitRun(i,stat);
  case EXT:
    return ExecuteInitExt(i,stat);
  case CALL:
    return ExecuteInitCall(i,stat);
  case IF:
    return ExecuteInitIf(i,stat);
  case SWITCH:
    return ExecuteInitSwitch(i,stat);
  case CASGN:
    WriteInitErr(stat,
                 "Incorrect statement type (constant assigned)"
                 " in initialization section");
    break;
  default:
    WriteInitErr(stat,"Incorrect statement type in initialization section");
    break;
  }
  /* all functions in the above switch should return only
     ok or stack_exceeded, handling all other errors internally. */
  return Proc_all_ok;
}

/* returns only ok or stack_exceeded messages */
static enum Proc_enum ExecuteInitStatements(struct Instance *i,
                                            struct StatementList *sl)
{
  unsigned c,length;
  struct gl_list_t *statements;
  struct Statement *stat;
  enum Proc_enum rval = Proc_all_ok;

  statements = GetList(sl);
  length = gl_length(statements);
  for(c = 1; c <= length && rval == Proc_all_ok; c++){
    stat = (struct Statement *)gl_fetch(statements,c);
    rval = ExecuteInitStatement(i,stat);
  }
  return rval;
}

/*
 * here's where we enforce stack limits (approximately)
 * this function returns only exceeded messages and ok messages.
 */
static enum Proc_enum ExecuteInitProcedure(struct Instance *i,
                                           struct InitProcedure *proc)
{
  struct for_table_t *OldForTable;
  enum Proc_enum rval = Proc_all_ok;

  g_proc.depth++;
  assert(i&&proc);
  if (g_proc.depth > g_proc.limit) {
    g_proc.depth--;
    return Proc_stack_exceeded_this_frame;
  }

  OldForTable = EvaluationForTable;
  EvaluationForTable = CreateForTable();
  rval = ExecuteInitStatements(i,ProcStatementList(proc));
  if (rval == Proc_stack_exceeded_this_frame) {
    FPRINTF(ASCERR,
      "Init function stack limit (%lu) reached. Possible recursion\n",
      g_proc.limit);
    rval = Proc_stack_exceeded;
  }
  if (rval == Proc_stack_exceeded) {
    FPRINTF(ASCERR,"(depth %lu) called in instance ",g_proc.depth);
    WriteInstanceName(ASCERR,i,NULL);
    FPRINTF(ASCERR,"\n");
  }
  DestroyForTable(EvaluationForTable);
  EvaluationForTable = OldForTable;
  g_proc.depth--;
  return rval;
}

/* returns overflow or ok. possibly either form of overflow */
static enum Proc_enum RealInitialize(struct Instance *context,
                                     struct Name *name)
{
  struct Name *instname=NULL;
  enum Proc_enum return_value = Proc_all_ok;
  struct Instance *ptr;
  enum find_errors err;
  struct InitProcedure *proc;
  struct gl_list_t *instances;
  unsigned long c,length;
  CONST char *procname=NULL;
  int previous_context = DeclarativeContext;

  DeclarativeContext = 1; /* set up for procedural processing */
  InstanceNamePart(name,&instname,&procname);
  if (procname!=NULL){
    if ((instances=FindInstances(context,instname,&err))){
      length = gl_length(instances);
      for(c=1;c<=length && return_value != Proc_stack_exceeded; c++){
	ptr = (struct Instance *)gl_fetch(instances,c);
	if ((proc = FindProcedure(ptr,procname))){
	  return_value = ExecuteInitProcedure(ptr,proc);
	} else {
	  return_value = Proc_proc_not_found;
        }
      }
      gl_destroy(instances);
    } else {			/* unable to find instances */
      return_value = Proc_instance_not_found;
    }
  } else {
    return_value = Proc_bad_name;
  }
  DeclarativeContext = previous_context;
  DestroyName(instname);
  return return_value;
}

enum Proc_enum Initialize(struct Instance *context, struct Name *name)
{
  enum Proc_enum rval;
  g_proc.depth = 0;
  rval = RealInitialize(context,name);
  if (rval == Proc_stack_exceeded || rval == Proc_stack_exceeded_this_frame) {
    FPRINTF(ASCERR,"Initialization stack overflow (possible recursion)\n");
    FPRINTF(ASCERR,"in user call to ");
    WriteName(ASCERR,name);
    FPRINTF(ASCERR," (depth 0) in instance ");
    WriteInstanceName(ASCERR,context,NULL);
    FPRINTF(ASCERR,"\n");
    g_proc.depth = 0;
    return Proc_stack_exceeded;
  }
/*    g_proc.depth = 0;
    return Proc_all_ok;    nope, return rval and DEAL WITH IT
*/
  return rval;
}

/*
 * This deals with initializations of the form:
 * RUN Type::procname; where Type is model or atom type,
 * and procname is a procedure defined within that type.
 * If the Type happened to have redefined a procedure from its
 * parent class, that procedure would be the one on its
 * procedure list and hence the one that would be invoked.
 *
 * returns only ok and stack_exceeded in the case of
 * correct arguments.
 */
static enum Proc_enum ClassAccessRealInitialize(struct Instance *context,
				     struct Name *class,
				     struct Name *name)
{
  enum Proc_enum return_value = Proc_all_ok;
  struct InitProcedure *proc;
  struct gl_list_t *plist;
  CONST char *procname;
  CONST char *typename;
  struct TypeDescription *desc,*conformable;
  int previous_context = DeclarativeContext;

  DeclarativeContext = 1; /* set up for procedural processing */

  typename = SimpleNameIdPtr(class);
  if (typename!=NULL) {
    desc = FindType(typename);
    if (desc!=NULL) {
      conformable = InstanceTypeDesc(context);
      if (MoreRefined(conformable,desc)) {
	plist = GetInitializationList(desc);
	if (plist) {
	  procname = SimpleNameIdPtr(name);
	  if (procname!=NULL) {
	    proc = SearchProcList(plist,procname);
	    if (proc) {
	      return_value = ExecuteInitProcedure(context,proc);
	    } else {
	      return_value = Proc_proc_not_found;
            }
	  } else {
	    return_value = Proc_illegal_name_use;
          }
	} else {
	  return_value = Proc_proc_not_found;
        }
      } else {
	return_value = Proc_illegal_type_use;
      }
    } else {
      return_value = Proc_type_not_found;
    }
  } else {
    return_value = Proc_illegal_name_use;
  }

  DeclarativeContext = previous_context;
  return return_value;
}

enum Proc_enum ClassAccessInitialize(struct Instance *context,
				     struct Name *class,
				     struct Name *name)
{
  /* in the fix, you must not miss ClassAccessInitialize */
  enum Proc_enum rval;
  g_proc.depth = 0;
  rval = ClassAccessRealInitialize(context,class,name);
  if (rval == Proc_stack_exceeded || rval == Proc_stack_exceeded_this_frame) {
    FPRINTF(ASCERR,"Initialization stack overflow (possible recursion)\n");
    FPRINTF(ASCERR,"in user call to ");
    WriteName(ASCERR,class);
    FPRINTF(ASCERR,"::");
    WriteName(ASCERR,name);
    FPRINTF(ASCERR," (depth 0) in instance ");
    WriteInstanceName(ASCERR,context,NULL);
    FPRINTF(ASCERR,"\n");
    g_proc.depth = 0;
    return Proc_stack_exceeded;
  }
  return rval;
}
