/*
 *  Procedure Data Structure Implementation
 *  by Tom Epperly
 *  Created: 1/10/90
 *  Version: $Revision: 1.14 $
 *  Version control file: $RCSfile: proc.c,v $
 *  Date last modified: $Date: 97/07/26 12:04:08 $
 *  Last modified by: $Author: rv2a $
 *
 *  This file is part of the Ascend Language Interpreter.
 *
 *  Copyright (C) 1990, 1993, 1994 Thomas Guthrie Epperly
 *
 *  The Ascend Language Interpreter is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  The Ascend Language Interpreter is distributed in hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139 USA.  Check the file named
 *  COPYING.
 *
 */

#include "utilities/ascConfig.h"
#include "utilities/ascMalloc.h"
#include "compiler/compiler.h"
#include "general/list.h"
#include "compiler/fractions.h"
#include "compiler/dimen.h"
#include "compiler/functype.h"
#include "compiler/types.h"
#include "compiler/stattypes.h"
#include "compiler/statement.h"
#include "compiler/slist.h"
#include "compiler/proc.h"

#define PMALLOC(x) \
  x = (struct InitProcedure *)ascmalloc(sizeof(struct InitProcedure))
#ifndef lint
static CONST char ProcedureRCSid[] = "$Id: proc.c,v 1.14 97/07/26 12:04:08 rv2a Exp $";
#endif

struct InitProcedure *CreateProcedure(CONST char *name,
				      struct StatementList *stats)
{
  register struct InitProcedure *result;
  PMALLOC(result);
  assert(result!=NULL);
  result->name = name;
  result->slist = stats;
  if (stats!=NULL) AddContext(stats,context_METH);
  return result;
}

struct InitProcedure *CopyProcedure(struct InitProcedure *p)
{
  register struct InitProcedure *result;
  assert(p!=NULL);
  PMALLOC(result);
  assert(result!=NULL);
  result->name = p->name;
  result->slist = CopyStatementList(p->slist);
  return result;
}

struct InitProcedure *CopyProcToModify(struct InitProcedure *p)
{
  register struct InitProcedure *result;
  assert(p!=NULL);
  PMALLOC(result);
  assert(result!=NULL);
  result->name = p->name;
  result->slist = CopyListToModify(p->slist);
  return result;
}

struct gl_list_t *MergeProcedureLists(struct gl_list_t *pl1,
                                       struct gl_list_t *pl2)
{
  register struct gl_list_t *list;
  register struct InitProcedure *proc;
  register unsigned long c,len;
  register unsigned long refproc;
  if (pl1==NULL) {
    return pl2;
  }
  list = gl_create(gl_length(pl1)+gl_safe_length(pl2));
  if (pl2!=NULL){
    len = gl_length(pl2);
    for(c=1;c<=len;c++){
      proc = (struct InitProcedure *)gl_fetch(pl2,c);
      gl_append_ptr(list,(VOIDPTR)proc);
    }                                 /* dont destroy the list yet */
  }
  len = gl_length(pl1);
  if (pl2!=NULL) {
    for(c=1;c<=len;c++) {
      proc = (struct InitProcedure *)gl_fetch(pl1,c);
      refproc = gl_search(pl2,proc,(CmpFunc)CmpProcs);
      /* refproc will be 0 if no match */
      if (refproc==0) {
        /* hence append to list */
        gl_append_ptr(list,(VOIDPTR)CopyProcedure(proc));
      }
    }
    gl_destroy(pl2);
  } else {
    for(c=1;c<=len;c++){
      proc = (struct InitProcedure *)gl_fetch(pl1,c);
      gl_append_ptr(list,(VOIDPTR)CopyProcedure(proc));
    }
  }
  gl_sort(list,(CmpFunc)CmpProcs);
  return list;
}

void DestroyProcedure(struct InitProcedure *p)
{
  if (p!=NULL){
    /* the following only destroys a reference to p->slist */
    DestroyStatementList(p->slist);
    ascfree((char *)p);
  }
}

void DestroyProcedureList(struct gl_list_t *pl)
{
  if (pl!=NULL){
    gl_iterate(pl,(void (*)(VOIDPTR))DestroyProcedure);
    gl_destroy(pl);
  }
}


CONST char *ProcNameF(CONST struct InitProcedure *p)
{
  assert(p!=NULL);
  return p->name;
}

struct StatementList *ProcStatementListF(CONST struct InitProcedure *p)
{
  assert(p!=NULL);
  return p->slist;
}

int CmpProcs(CONST struct InitProcedure *p1, CONST struct InitProcedure *p2)
{
  assert(p1 && p2);
  return strcmp(p1->name,p2->name);
}


int CompareProcedureLists(struct gl_list_t *pl1, struct gl_list_t *pl2,
                          unsigned long *n) 
{
  CONST struct InitProcedure *proc1, *proc2;
  CONST struct StatementList *sl1, *sl2;  
  unsigned long c,len, len1, len2, diff;

  assert (n != NULL);

  /*  compare ptrs */
  if (pl1==pl2) {
    return 0;
  }

  /* check for nulls */
  if ( (pl1==NULL) || (pl2==NULL) ) {
    *n = 1;
    return 1;
  }

  /* length  of each procedure list */
  len1 = gl_length(pl1);
  len2 = gl_length(pl2);

  /* 
   * if len1 and len2 are different from each other , we will return 1, 
   * but we still need to find the index of the first different procedure. 
   * So, get the number of elements of the shorthest list and compare 
   * that number of procedures in the lists
   */
  if (len1 == len2) {
    len = len1;
  } else {
    if (len1 < len2) {
      len = len1;
    } else {
      len = len2;
    }
  }

  /* analyze the list of procedures */
  for(c=1;c<=len;c++){
    proc1 = (struct InitProcedure *)gl_fetch(pl1,c);
    proc2 = (struct InitProcedure *)gl_fetch(pl2,c);  

    /* 
     * compare names 
     */ 
    if (CmpProcs(proc1,proc2)) {
      *n = c;
      return 1;
    }
    /* 
     * compare statement lists
     * if lists are different, diff will contain the index of the first
     * different statement. Use it if necessary
     */
    sl1 = ProcStatementList(proc1);
    sl2 = ProcStatementList(proc2); 
    if (CompareStatementLists(sl1,sl2,&diff)) {
      *n = c;
      return 1;
    }
  }

  /*
   * If we get here and the number of elements in the list are equal,
   * then the procedures are equal,
   * else the index of the first different procedure is len + 1
   */
  if (len1 == len2) {
    return 0;
  } else {
    *n = len+1;
    return 1;
  }
}
